/*
  Copyright 1996-2002
  Simon Whiteside

  $Id: Interpreter.java,v 1.44 2002/02/13 17:52:46 sdw Exp $
*/
package simkin;
import java.io.*;
import java.util.Vector;
import java.util.Hashtable;
import java.lang.reflect.Method;
import java.lang.reflect.Field;

/**
 * This class represents the Simkin interpreter. The interpreter can load, parse and interpret Simkin code.
 * <p>The Interpreter maintains a list of global variables. This maps names to objects, which are available to all Simkin scripts.
 * <p>The interpreter itself is a global object with the name "Interpreter".
 * <p>The Interpreter implements the {@link simkin.Executable Executable} interface, supports the single field "tracing" which enables/disables tracing of method calls
 * <p>New from Simkin 1.16: by default a different instance of the Interpreter is created for each thread. This can be overriden with the {@link simkin.Interpreter#setInterpreterPerThread setInterpreterPerThread} method
 */
public class Interpreter extends ExecutableRoot {
    /**
     * this flag indicates if there is one interpreter for the process, or one per thread. The default is true
     */
    static boolean g_InterpreterPerThread=true;
    /**
     * this is the interpreter used by objects when a call into simkin is required and there is one interpreter per process
     */
    static Interpreter g_Interpreter;
    /**
     * this is used to find the correct interpreter for the calling thread
     */
    static Hashtable g_Interpreters=new Hashtable();
    /**
     * this string holds the location of the current method
     */
    //    String m_Location;
    /** This holds the value of the current line number being executed */
    //    int m_LineNum=0;
    /**
     * this is the name of the interpreter global variable
     */
    static String s_Interpreter="Interpreter";
    /**
     * this is the name of the Interpreter's tracing variable, if true, the interpreter prints out tracing information
     */
    static String s_Tracing="tracing";
    /**
     * This table holds the global variables for this interpreter
     */
    Hashtable m_GlobalVars=new Hashtable();
    /**
     * if this variable is true, the interpreter outputs information about the execution of the code (method calls and return values)
     */
    boolean m_Tracing=false;
    /**
     * this variable sets tracing on for the functionality of the interpreter
     */
    static final boolean g_TraceExecution=false;
    /**
     * the null object
     */
    static public final Null g_null=new Null();    
    /** true object */
    static public final Boolean g_True=new Boolean(true);
    /** false object */
    static public final Boolean g_False=new Boolean(false);
    /** zero object */
    static public final Integer g_Zero=new Integer(0);
    /**
     * Object to receive trace message for this interpreter object
     */
    TraceCallback m_TraceCallback=null;
    /** Object to receive statement notifications */
    StatementStepper m_StatementStepper=null;

    /**
     * Default constructor - adds a global variable pointing to this interpreter
     */
    public Interpreter(){
	addGlobalVariable(s_Interpreter,this);
    }
    /**
     * Used to set the "tracing" variable - for example in Simkin:
     * <p> Interpreter.tracing=true;
     */
    public void setValue(String field_name,String attrib,Object value) throws RuntimeException,FieldNotSupportedException{
	if (field_name.equals(s_Tracing)){
	    m_Tracing=boolValue(value);
	}else
	    throw new FieldNotSupportedException("");
    }
    /**
     * The interpreter does not expose any methods to Simkin scripts
     */
    public Object method(String method_name,Object[] args) throws RuntimeException,ParseException,MethodNotSupportedException{
	throw new MethodNotSupportedException("");
    }
    /**
     * This method converts the given value into a boolean value
     * <p>If the object is an instance of Boolean, the boolean value is returned, otherwise toString() is called, and
     * compared with "true"
     * @return the boolean equivalent of the object
     */
    public static boolean boolValue(Object value){
	boolean ret=false;
	if (value!=null){
	    if (value instanceof Boolean)
		ret=((Boolean)value).booleanValue();
	    else
		if (value.toString().equals(NodeTypes.s_True))
		    ret=true;
	}
	return ret;
    }
    /**
     * This method converts the given value into an integer value
     * <p>If the object is an instance of Integer, the integer value is returned, otherwise toString() is called, and
     * converted to an integer
     * @return the integer equivalent of the object
     */
    public static int intValue(Object value){
	int ret=0;
	if (value instanceof Integer)
	    ret=((Integer)value).intValue();
	else{
	    try{
		ret=(int)Math.floor(floatValue(value));
	    }catch(Exception e){
	    }
	}
	return ret;
    }
    /**
     * This method converts the given value into a double value
     * <p>If the object is an instance of Double, the integer value is returned, otherwise toString() is called, and
     * converted to a double
     * @return the double equivalent of the object
     */
    public static double floatValue(Object value){
	double ret=0;
	if (value instanceof Double)
	    ret=((Double)value).doubleValue();
	else{
	    try{
		ret=new Double(value.toString()).doubleValue();
	    }catch(Exception e){
	    }
	}
	return ret;
    }
    /** Returns null - this is not a collection */
    public Object getValueAt(Object index,String attrib_name) throws RuntimeException{
	return null;
    }
    /** Does nothing - this is not a collection */
    public void setValueAt(Object index,String attrib_name,Object value) throws RuntimeException{
    }
    /**
     * Returns the value of the "tracing" variable - e.g. in Simkin:
     * <p> tracing=Interpreter.tracing;
     */
    public Object getValue(String field_name,String attrib) throws RuntimeException,FieldNotSupportedException{
	Object o=null;
	boolean bRet=false;
	if (field_name.equals(s_Tracing))
	    o=new Boolean(m_Tracing);
	else
	    throw new FieldNotSupportedException("");
	return o;
    }
    /**
     * Called to execute a statement
     * @return a return object
     */
    Object executeStat(Context ctxt,Object obj,Hashtable var,StatNode stat) throws ParseException, RuntimeException{
	if (g_TraceExecution) trace("executeStat "+NodeTypes.g_NodeNames[stat.getType()]);
	Object ret=null;
	ctxt.m_LineNum=stat.getLineNumber();
	int stat_type=stat.getType();
	boolean stop=false;
	if (m_StatementStepper!=null)
	    stop=(m_StatementStepper.statementExecuted(ctxt.m_Location,ctxt.m_LineNum,obj,var,stat_type)==false);
	if (stop==false){
	    switch(stat_type){
	    case NodeTypes.s_StatList:
		ret=executeStats(ctxt,obj,var,(StatListNode)stat);
		break;
	    case NodeTypes.s_Method:
		evalMethod(ctxt,obj,var,((MethodNode)stat).m_Ids);
		break;
	    case NodeTypes.s_if:
		ret=executeIfStat(ctxt,obj,var,(IfNode)stat);
		break;
	    case NodeTypes.s_For:
		ret=executeForStat(ctxt,obj,var,(ForNode)stat);
		break;
	    case NodeTypes.s_ForEach:
		ret=executeForEachStat(ctxt,obj,var,(ForEachNode)stat);
		break;
	    case NodeTypes.s_while:
		ret=executeWhileStat(ctxt,obj,var,(WhileNode)stat);
		break;
	    case NodeTypes.s_Assign:
		executeAssignStat(ctxt,obj,var,(AssignNode)stat);
		break;
	    case NodeTypes.s_return:
		ret=executeReturnStat(ctxt,obj,var,(ReturnNode)stat);
		break;
	    case NodeTypes.s_switch:
		ret=executeSwitchStat(ctxt,obj,var,(SwitchNode)stat);
		break;
	    }
	}else
	    ret=g_null;
	return ret;
    }
    /**
     * Called to execute a block of statements
     * @return a return object
     */
    Object executeStats(Context ctxt,Object obj,Hashtable var,StatListNode n) throws ParseException, RuntimeException{
	if (g_TraceExecution) trace("executeStats "+n.m_Stats.size());
	Object ret=null;
	boolean stop=false;
	if (m_StatementStepper!=null)
	    stop=(m_StatementStepper.compoundStatementExecuted(ctxt.m_Location,ctxt.m_LineNum,obj,var)==false);
	if (stop==false){
	    for (int i=0;i<n.m_Stats.size();i++){
		StatNode stat=(StatNode)(n.m_Stats.elementAt(i));
		ret=executeStat(ctxt,obj,var,stat);
		if (ret!=null)
		    break;
	    }
	}else
	    ret=g_null;
	return ret;
    }
    /**
     * Adds a local variable to the current list of local variables
     */
    void addLocalVariable(Hashtable var,String name,Object value){
	if (g_TraceExecution) trace("Adding local variable:" +name);
	var.put(name,value);
    }
    /**
     * looks for the given global variable.
     * @param name - the name of the global variable
     * @return the object found, or null if the global variable is undefined
     */
    public  Object findGlobalVariable(String name){
	synchronized(m_GlobalVars) {
	    return m_GlobalVars.get(name);
	}
    }
    /**
     * Adds a global variable to the current list
     * @param name - the name of the global variable, if one already exists it is replaced
     * @param r - the object to be added to the list
     */
    public  void addGlobalVariable(String name,Object r){
	synchronized(m_GlobalVars) {
	    m_GlobalVars.put(name,r);
	}
    }
    /**
     * Removes a global variable from the current list
     * @param name - the name of the global variable to be removed
     */
    public void removeGlobalVariable(String name){
	synchronized(m_GlobalVars) {
	    m_GlobalVars.remove(name);
	}
    }
    /**
     * Evaluates a method and returns the result
     * @param obj the object owning the code
     * @param var the current local variable list
     * @param ids the id list
     * @return the value from the method call
     */
    Object evalMethod(Context ctxt,Object obj,Hashtable var,IdListNode ids) throws ParseException, RuntimeException{
	Object ret=null;
	IdNode idNode=(IdNode)ids.m_Ids.elementAt(ids.m_Ids.size()-1);
	String method_name=idNode.m_Name;
	if (ids.m_Ids.size()==1){
	    ret=makeMethodCall(ctxt,obj,var,obj,method_name,idNode.m_ArrayIndex,ids.m_Attrib,idNode.m_Exprs);
	}else{
	    Object called_obj=followIdList(ctxt,obj,var,ids);
	    ret=makeMethodCall(ctxt,obj,var,called_obj,method_name,idNode.m_ArrayIndex,ids.m_Attrib,idNode.m_Exprs);
	}
	return ret;
    }
    /**
     * executes a return statement by copying the expression's value into the return variable
     * @return the result of the result expression
     */
    Object executeReturnStat(Context ctxt,Object obj,Hashtable var,ReturnNode n) throws ParseException, RuntimeException{
	return evaluate(ctxt,obj,var,n.m_Expr);
    }
    /**
     * executes a while statement by repeatedly executing the block of statements until the test expression is false
     * @return the return object
     */
    Object executeWhileStat(Context ctxt,Object obj,Hashtable var,WhileNode n) throws ParseException, RuntimeException{
	Object ret=null;
	// loop until an enclosing statement is a return statement
	while(ret==null){
	    Object expr=evaluate(ctxt,obj,var,n.m_Expr);
	    // or until the loop expression is false
	    if (boolValue(expr)){
		if (n.m_Stats!=null)
		    ret=executeStat(ctxt,obj,var,n.m_Stats);
	    }else
		break;
	}
	return ret;
    }
    /**
     * executes a for each statement by obtaining an ExecutableIterator object and iterating over it
     * @return the return object
     */
    Object executeForEachStat(Context ctxt,Object obj,Hashtable var,ForEachNode n) throws ParseException, RuntimeException{
	Object ret=null;
	String checked_id=checkIndirectId(ctxt,obj,var,n.m_Id.m_Name);
	Object expr=evaluate(ctxt,obj,var,n.m_Expr);
	if (expr instanceof Executable){
	    ExecutableIterator iterator=null;
	    if (n.m_Qualifier!=null)
		iterator=((Executable)expr).createIterator(n.m_Qualifier.m_Name);
	    else
		iterator=((Executable)expr).createIterator();
	    if (iterator!=null){
		Object value;
		while (ret==null){
		    value=iterator.next();
		    if (value==null)
			break;
		    else{
			addLocalVariable(var,checked_id,value);
			if (n.m_Stats!=null){
			    ret=executeStats(ctxt,obj,var,n.m_Stats);
			}
		    }
		}
	    }else
		runtimeException(ctxt,Messages.CouldNotCreateIterator);
	}else
	    runtimeException(ctxt,Messages.CannotApplyForEach);
	return ret;
    }

    /**
     * executes a for statement by repeatedly executing the block of statements for the range given in the statement
     * @return the return object
     */
    Object executeForStat(Context ctxt,Object obj,Hashtable var,ForNode n) throws ParseException, RuntimeException{
	Object ret=null;
	String checked_id=checkIndirectId(ctxt,obj,var,n.m_Id.m_Name);
	Object start_expr=evaluate(ctxt,obj,var,n.m_StartExpr);
	int start_value=intValue(start_expr);
	Object end_expr=evaluate(ctxt,obj,var,n.m_EndExpr);
	int end_value=intValue(end_expr);
	int step_value=1;
	if (n.m_StepExpr!=null){
	    Object step_expr=evaluate(ctxt,obj,var,n.m_StepExpr);
	    step_value=intValue(step_expr);
	}
	if (end_value!=start_value){
	    if (step_value>0){
		if (end_value>start_value){
		    for (int i=start_value;i<end_value;i+=step_value){
			addLocalVariable(var,checked_id,new Integer(i));
			if (n.m_Stats!=null){
			    ret=executeStats(ctxt,obj,var,n.m_Stats);
			    if (ret!=null)
				break;
			}
		    }
		}else
		    runtimeException(ctxt,Messages.EndNotGreaterThanStart);
	    }else{
		if (step_value<0){
		    if (start_value>end_value){
			for (int i=start_value;i>end_value;i+=step_value){
			    addLocalVariable(var,checked_id,new Integer(i));
			    if (n.m_Stats!=null){
				ret=executeStats(ctxt,obj,var,n.m_Stats);
				if (ret!=null)
				    break;
			    }
			}
		    }else
			runtimeException(ctxt,Messages.StartNotGreaterThanEnd);
		}else
		    runtimeException(ctxt,Messages.ZeroStep);
	    }
	}// else - do nothing
	return ret;
    }
    /**
     * executes a switch statement by evaluating an expression and testing it for equality against the case statements
     * @return the return object
     */
    Object executeSwitchStat(Context ctxt,Object obj,Hashtable var,SwitchNode n) throws ParseException, RuntimeException{
	Object ret=null;
	Object expr=evaluate(ctxt,obj,var,n.m_Expr);
	CaseNode caseFound=null;
	for (int i=0;i<n.m_Cases.m_Cases.size();i++){
	    if (g_TraceExecution) trace("checking case "+i);
	    CaseNode caseNode=(CaseNode)(n.m_Cases.m_Cases.elementAt(i));
	    Object testExpr=evaluate(ctxt,obj,var,caseNode.m_Expr);
	    if ((expr==null && testExpr==null) || (expr!=null && expr.equals(testExpr))){
		caseFound=caseNode;
		break;
	    }
	}
	if (caseFound!=null){
	    ret=executeStats(ctxt,obj,var,caseFound.m_Stats);
	}else if (n.m_DefaultCase!=null)
	    ret=executeStats(ctxt,obj,var,n.m_DefaultCase.m_Stats);
	return ret;
    }
    /**
     * Executes an if..else statement
     * @return the return object
     */
    Object executeIfStat(Context ctxt,Object obj,Hashtable var,IfNode n) throws ParseException, RuntimeException{
	Object ret=null;
	Object rExpr=evaluate(ctxt,obj,var,n.m_Expr);
	if (boolValue(rExpr)){
	    if (n.m_Stats!=null)
		ret=executeStat(ctxt,obj,var,n.m_Stats);
	}else
	    if (n.m_Else!=null)		
		ret=executeStat(ctxt,obj,var,n.m_Else);
	return ret;
    }
    /**
     * Evaluates a series of argument expressions for a method call
     * @param obj - the object owning the script
     * @param var - the current local variables
     * @param n - the node containing the list of expressions
     * @param args - the array of arguments to be filled up
     */
    void evalArgs(Context ctxt,Object obj,Hashtable var,ExprListNode n,Object [] args) throws ParseException, RuntimeException{
	if (n!=null && n.m_Exprs!=null)
	    for (int i=0;i<n.m_Exprs.size();i++){
		ExprNode parg=(ExprNode)(n.m_Exprs.elementAt(i));
		args[i]=evaluate(ctxt,obj,var,parg);
	    }
    }
    /**
     * Checks whether an identifier is prefixed by "@". If it is, the id is first evaluated and then a variable with the value's name returns as the new id
     * @return the name, once the indirection has been resolved
     */
    String checkIndirectId(Context ctxt,Object obj,Hashtable var,String name) throws ParseException{
	if (name.length()>0 && name.charAt(0)=='@'){
	    name=name.substring(1,name.length());
	    Object new_name=findValue(ctxt,obj,var,name,null,null);
	    if (new_name!=null){
		name=new_name.toString();
	    }
	}
	return name;
    }	
    /**
     * searches for a named value. Looking first for built-in names, then within the given local variable table, then within the given object, and finally within the global variables
     * @param obj the object whose fields will be searched
     * @param var the local variable list
     * @param name the name of the value being sought
     * @param attrib the name of the attribute being sought (might be null)
     * @return the object found
     */
    Object findValue(Context ctxt,Object obj,Hashtable var,String name,ExprNode array_index,String attrib) throws ParseException{
	Object ret=null;
	name=checkIndirectId(ctxt,obj,var,name);
	if (name.length()>0){
	    // the name is prefixed with "self"
	    if (name.equals(NodeTypes.s_null)){
		ret=g_null;
	    }else if (name.equals(NodeTypes.s_Self)){
		if (attrib!=null || array_index!=null){
		    if (array_index!=null)
			ret=extractArrayValue(ctxt,obj,var,obj,array_index,attrib);
		    else
			ret=extractValue(ctxt,obj,var,obj,null,attrib);
		}else
		    ret=obj;
		// the value is "true"
	    }else if (name.equals(NodeTypes.s_True))
		ret=g_True;
	    // the value is "false"
	    else if (name.equals(NodeTypes.s_False))
		ret=g_False;
	    else{
		// otherwise first check in the local variables
		ret=var.get(name);
		if (ret==null){
		    // then in the instance data for the object
		    try{
			if (g_TraceExecution) trace("looking for instance var:"+name);
			if (obj!=null){
			    if (array_index!=null)
				ret=extractFieldArrayValue(ctxt,obj,var,obj,name,array_index,attrib);
			    else
				ret=extractValue(ctxt,obj,var,obj,name,attrib);
			}else{
			    synchronized(m_GlobalVars) {
				ret=extractValue(ctxt,obj,var,m_GlobalVars.get(name),null,attrib);
			    }
			    if (ret==null)
				runtimeException(ctxt,Messages.formatMessage(Messages.FieldNotFound,name));
			    else
				// handle the array or attribute case
				if (attrib!=null || array_index!=null){
				    if (array_index!=null)
					ret=extractArrayValue(ctxt,obj,var,ret,array_index,attrib);
				    else
					ret=extractValue(ctxt,obj,var,ret,null,attrib);
				}
			}
		    }catch(Exception e){
			if (g_TraceExecution) trace("looking for global var:"+name);
			// finally in the global variables list
			ret=extractValue(ctxt,obj,var,m_GlobalVars.get(name),null,attrib);
			if (ret==null)
			    runtimeException(ctxt,Messages.formatMessage(Messages.FieldNotFound,name)+" ("+e.toString()+")");
			else
			    // handle the array or attribute case
			    if (attrib!=null || array_index!=null){
				if (array_index!=null)
				    ret=extractArrayValue(ctxt,obj,var,ret,array_index,attrib);
				else
				    ret=extractValue(ctxt,obj,var,ret,null,attrib);
			    }
		    }
		}else{
		    if (attrib!=null || array_index!=null){
			if (array_index!=null)
			    ret=extractArrayValue(ctxt,obj,var,ret,array_index,attrib);
			else
			    ret=extractValue(ctxt,obj,var,ret,null,attrib);
		    }
		}
	    }
	}	
	return ret;
    }
    /** this method dereferences an instance variable of an object, and then applies an array index to the result
     * @param obj the object owning the context of the current code
     * @param var the local variables for the current contaxt
     * @param called_obj the object owning the field to be dereferenced
     * @param field_name the name of the field
     * @param array_index an expression denoting the array index
     * @param attrib an optional attribute (cann be null)
     */
    Object extractFieldArrayValue(Context ctxt,Object obj,Hashtable var,Object called_obj,String field_name,ExprNode array_index,String attrib) throws ParseException{
	if (g_TraceExecution) trace("extractFieldArrayValue:"+field_name);
	Object ret=extractValue(ctxt,obj,var,called_obj,field_name,null);
	ret=extractArrayValue(ctxt,obj,var,ret,array_index,attrib);
	return ret;
    }
    /**
     * This method extracts a value from an array object
     * @param obj - the owner of the current context
     * @param var - local variables
     * @param called_obj - the owner of the value
     * @param array_index - an expression indicating an array index 
     * @param attrib - the name of an attribute (can be null)
     */
    Object extractArrayValue(Context ctxt,Object obj,Hashtable var,Object called_obj,ExprNode array_index,String attrib) throws ParseException{
	if (g_TraceExecution) trace("extractArrayValue:");
	Object ret=called_obj;
	Object array_object=evaluate(ctxt,obj,var,array_index);
	if (ret instanceof Executable)
	    ret=((Executable)ret).getValueAt(array_object,attrib);
	else if (ret.getClass().isArray()){
	    if (attrib==null){
		Object[] array=(Object[])ret;
		ret=array[intValue(array_object)];
	    }else
		runtimeException(ctxt,Messages.NoAttributesOnArrays);
	}else
	    runtimeException(ctxt,Messages.ArrayIndexNotOnExecutable);
	if (ret==null)
	    ret=g_null;
	return ret;
    }
    /**
     * This method extracts a value from an object
     * @param obj - the owner of the current context
     * @param var - local variables
     * @param called_obj - the owner of the value
     * @param name - the name of the field (can be null)
     * @param attrib - the name of an attribute (can be null)
     */
    Object extractValue(Context ctxt,Object obj,Hashtable var,Object called_obj,String name,String attrib) throws ParseException{
	if (name!=null)
	    name=checkIndirectId(ctxt,obj,var,name);
	if (g_TraceExecution) trace("extractValue:"+name+" "+attrib);
	Object ret=called_obj;
	if (attrib!=null){
	    // if an attribute is passed, this only works with Executable-type objects
	    if (called_obj instanceof Executable){
		try{
		    ret=((Executable)called_obj).getValue(name,attrib);
		}catch(Exception e){
		    runtimeException(ctxt,Messages.formatMessage(Messages.AttributeNotFound,attrib)+" ("+e.toString()+")");
		}
	    }else{
		runtimeException(ctxt,Messages.AttributeNotOnExecutable);
	    }
	}else{
	    if (name!=null){
		if (called_obj instanceof Executable){
		    try{
			ret=((Executable)called_obj).getValue(name,attrib);
		    }catch(Exception e){
			runtimeException(ctxt,Messages.formatMessage(Messages.FieldNotFound,name)+" ("+e.toString()+")");
		    }
		}else{
		    try{
			ret=reflectiveGetValue(ctxt,called_obj,name);
		    }catch(Exception e){
			//			trace(e);
			runtimeException(ctxt,Messages.formatMessage(Messages.FieldNotFound,name)+" ("+e.toString()+")");
		    }
		}
	    }
	}
	if (ret==null)
	    ret=g_null;
	return ret;
    }
    /**
     * this method assigns the value of an expression to a field in another object
     * @param obj the object in which context the method is executed
     * @param var the local variable list
     * @param n the portion of the parse tree for the assignment statement
     */
    void executeAssignStat(Context ctxt,Object obj,Hashtable var,AssignNode n) throws ParseException, RuntimeException{
	Object value=evaluate(ctxt,obj,var,n.m_Expr);
	IdNode idNode=(IdNode)(n.m_Ids.m_Ids.elementAt(n.m_Ids.m_Ids.size()-1));
	String name=checkIndirectId(ctxt,obj,var,idNode.m_Name);
	if (n.m_Ids.m_Ids.size()==1){
	    if (name.equals(NodeTypes.s_Self)){
		try{
		    if (idNode.m_ArrayIndex!=null)
			insertArrayValue(ctxt,obj,var,obj,idNode.m_ArrayIndex,n.m_Attrib,value);
		    else
			insertValue(ctxt,obj,var,obj,null,n.m_Attrib,value);
		}catch(Exception e){
		    runtimeException(ctxt,Messages.formatMessage(Messages.CouldNotSetAttributeOnClass,n.m_Attrib,obj.getClass().getName())+" ("+e.toString()+")");
		}
	    }else{
		Object caller=obj;
		try{
		    if (idNode.m_ArrayIndex!=null){
			caller=findValue(ctxt,obj,var,name,null,null);
			insertArrayValue(ctxt,obj,var,caller,idNode.m_ArrayIndex,n.m_Attrib,value);
		    }else{
			if (n.m_Attrib!=null){
			    insertValue(ctxt,obj,var,obj,name,n.m_Attrib,value);
			}else{
			    insertValue(ctxt,obj,var,obj,name,n.m_Attrib,value);
			}
		    }
		}catch(FieldNotSupportedException e){
		    addLocalVariable(var,name,value);
		}catch(Exception e){
		    //		    trace(e);
		    runtimeException(ctxt,Messages.formatMessage(Messages.CouldNotSetFieldOnClass,name,n.m_Attrib,caller.getClass().getName())+" ("+e.toString()+")");
		}
	    }
	}else{
	    Object called_obj=followIdList(ctxt,obj,var,n.m_Ids);
	    try{
		if(idNode.m_ArrayIndex!=null){
		    called_obj=extractValue(ctxt,obj,var,called_obj,name,null);
		    insertArrayValue(ctxt,obj,var,called_obj,idNode.m_ArrayIndex,n.m_Attrib,value);
		}else
		    insertValue(ctxt,obj,var,called_obj,name,n.m_Attrib,value);
	    }catch(Exception e){
		//		trace(e);
		runtimeException(ctxt,Messages.formatMessage(Messages.CouldNotSetFieldOnClass,name,n.m_Attrib,called_obj.getClass().getName())+" ("+e.toString()+")");
	    }
	}
    }
    /**
     * This method sets a value into an object
     * @param obj - the object owning the current context
     * @param var - the current local variable list
     * @param called_obj - the object owning the value
     * @param name - the name of the field
     * @param attrib - the attribute name
     * @param value - the value to be set
     */
    void insertValue(Context ctxt,Object obj,Hashtable var,Object called_obj,String name,String attrib,Object value) throws FieldNotSupportedException,ParseException{
	if (attrib!=null){
	    // this is the situation where the attribute is non-null
	    if (called_obj instanceof Executable){
		((Executable)called_obj).setValue(name,attrib,value);
	    }else{
		runtimeException(ctxt,Messages.AttributeNotOnExecutable);
	    }
	}else{
	    if (called_obj instanceof Executable)
		((Executable)called_obj).setValue(name,attrib,value);
	    else
		reflectiveSetValue(ctxt,called_obj,name,value);
	}
    }
    /**
     * This method sets a value into an array object
     * @param obj - the object owning the current context
     * @param var - the current local variable list
     * @param called_obj - the object owning the value
     * @param attrib - the attribute name
     * @param value - the value to be set
     */
    void insertArrayValue(Context ctxt,Object obj,Hashtable var,Object called_obj,ExprNode array_index,String attrib,Object value) throws FieldNotSupportedException,ParseException{
	Object array_object=evaluate(ctxt,obj,var,array_index);
	if (called_obj instanceof Executable)
	    ((Executable)called_obj).setValueAt(array_object,attrib,value);
	else if (called_obj.getClass().isArray()){
	    if (attrib==null){
		Object[] array=(Object[])called_obj;
		array[intValue(array_object)]=value;
	    }else
		runtimeException(ctxt,Messages.AttributeNotOnExecutable);
	}else{
	    runtimeException(ctxt,Messages.ArrayIndexNotOnExecutable);
	}
    }
    /**
     * This method follows an id list to the penultimate id, attempting to
     * locate objects
     * @param obj - owner of this method
     * @param var - current local variables
     * @param idlist - list of id's
     * @return object - the object found by following the list
     */
    Object followIdList(Context ctxt,Object obj,Hashtable var,IdListNode idList) throws ParseException{
	IdNode idNode=(IdNode)idList.m_Ids.elementAt(0);
	String name=checkIndirectId(ctxt,obj,var,idNode.m_Name);
	Object curr_obj=null;
	if (idNode.m_Exprs==null)
	    curr_obj=findValue(ctxt,obj,var,name,idNode.m_ArrayIndex,null);
	else
	    curr_obj=makeMethodCall(ctxt,obj,var,obj,name,idNode.m_ArrayIndex,null,idNode.m_Exprs);
	for (int i=1;i<idList.m_Ids.size()-1;i++){
	    idNode=(IdNode)idList.m_Ids.elementAt(i);
	    name=checkIndirectId(ctxt,obj,var,idNode.m_Name);
	    if (idNode.m_Exprs==null){
		if (idNode.m_ArrayIndex!=null)
		    curr_obj=extractFieldArrayValue(ctxt,obj,var,curr_obj,name,idNode.m_ArrayIndex,null);
		else
		    curr_obj=extractValue(ctxt,obj,var,curr_obj,name,null);
	    }else
		curr_obj=makeMethodCall(ctxt,obj,var,curr_obj,name,idNode.m_ArrayIndex,null,idNode.m_Exprs);
	}
	return curr_obj;
    }
    /**
     * This method calls a method on the given object
     * @param obj - the owner of the current context
     * @param var - the current local variables
     * @param called_obj - the object whose method is to be called
     * @param name - the name of the method
     * @param exprs - the expression list to be passed as parameters - these will be evaluated
     */
    Object makeMethodCall(Context ctxt,Object obj,Hashtable var,Object called_obj,String name,ExprNode array_index,String attrib,ExprListNode exprs) throws ParseException{
	String method_name=checkIndirectId(ctxt,obj,var,name);
	Object [] args=null;
	Object ret=null;
	if (exprs!=null && exprs.m_Exprs.size()>0)
	    args=new Object[exprs.m_Exprs.size()];
	evalArgs(ctxt,obj,var,exprs,args);
	if (m_Tracing)  traceString("Interpreter: "+called_obj.getClass().getName()+"::"+method_name+"()\n");
	try{
	    if (called_obj instanceof Executable){
		ret=((Executable)called_obj).method(method_name,args);	
	    }else
		ret=reflectiveMethodCall(ctxt,called_obj,method_name,args);
	    // check if there's an array index after this method - if so, dereference the object
	    if (array_index!=null){
		Object array_object=evaluate(ctxt,obj,var,array_index);
		if (ret instanceof Executable)
		    ret=((Executable)ret).getValueAt(array_object,attrib);
		else if (ret.getClass().isArray()){
		    Object[] array=(Object[])ret;
		    ret=array[intValue(array_object)];
		}else
		    runtimeException(ctxt,Messages.ArrayIndexNotOnExecutable);
	    }else
		if (attrib!=null){
		    // extract value here
		}
	}catch(MethodNotSupportedException e){
	    runtimeException(ctxt,Messages.formatMessage(Messages.CouldNotCallMethodClass,method_name,called_obj.getClass().getName())+" ("+e.toString()+")");
	}catch(Exception e){
	    //	    trace(e);
	    runtimeException(ctxt,e.toString());
	}
	if (m_Tracing)
	    traceString("Interpreter: "+method_name+"() returns "+(ret!=null?ret.toString():"null")+"\n");
	if (ret==null)
	    ret=g_null;
	return ret;
    }
    /** 
     * This method evaluates an expression
     * @param obj the object in whose context the expression is evaluated
     * @param var the local variable list
     * @param n the portion of the parse tree for this expression
     * @return the value of the expression
     */
    Object evaluate(Context ctxt,Object obj,Hashtable var,ExprNode n) throws ParseException, RuntimeException{
	Object ret=null;
	Object other=null;
	if (g_TraceExecution) trace("evaluate "+NodeTypes.g_NodeNames[n.getType()]);
	switch(n.getType()){
	case NodeTypes.s_IdList:{
	    IdListNode fnode=(IdListNode)n;
	    IdNode idNode=(IdNode)fnode.m_Ids.elementAt(fnode.m_Ids.size()-1);
	    String name=idNode.m_Name;
	    if (g_TraceExecution) trace("evaluate "+NodeTypes.g_NodeNames[n.getType()]+" name \""+name+"\"");
	    if (fnode.m_Ids.size()==1){
		if (idNode.m_Exprs==null)
		    ret=findValue(ctxt,obj,var,name,idNode.m_ArrayIndex,fnode.m_Attrib);
		else
		    ret=makeMethodCall(ctxt,obj,var,obj,name,idNode.m_ArrayIndex,fnode.m_Attrib,idNode.m_Exprs);
	    }else{
		name=checkIndirectId(ctxt,obj,var,name);
		Object called_obj=followIdList(ctxt,obj,var,fnode);
		if (idNode.m_Exprs==null){
		    if (idNode.m_ArrayIndex!=null)
			ret=extractFieldArrayValue(ctxt,obj,var,called_obj,name,idNode.m_ArrayIndex,fnode.m_Attrib);
		    else
			ret=extractValue(ctxt,obj,var,called_obj,name,fnode.m_Attrib);
		}else{
		    ret=makeMethodCall(ctxt,obj,var,called_obj,name,idNode.m_ArrayIndex,fnode.m_Attrib,idNode.m_Exprs);
		}
	    }
	    break;
	}
	case NodeTypes.s_String:
	    ret=((LiteralNode)n).m_Value;
	    break;
	case NodeTypes.s_Character:
	    ret=new Character(((LiteralNode)n).m_Value.charAt(0));
	    break;
	case NodeTypes.s_Integer:
	    try{
		ret=new Integer(Integer.parseInt(((LiteralNode)n).m_Value));
	    }catch(Exception e){
		ret=g_Zero;
	    }
	    break;
	case NodeTypes.s_Float:
	    ret=new Double(((LiteralNode)n).m_Value);
	    break;
	default:{
		OpNode o=(OpNode)n;
		Object item1=evaluate(ctxt,obj,var,o.m_Expr1);
		switch(n.getType()){
		case NodeTypes.s_not:
		    if (boolValue(item1))
			ret=g_False;
		    else
			ret=g_True;
		    break;
		case NodeTypes.s_and:
		    if (boolValue(item1) && boolValue(evaluate(ctxt,obj,var,o.m_Expr2)))
			ret=g_True;
		    else
			ret=g_False;
		    break;
		case NodeTypes.s_or:
		    if (boolValue(item1) || boolValue(evaluate(ctxt,obj,var,o.m_Expr2)))
			ret=g_True;
		    else
			ret=g_False;
		    break;
		case NodeTypes.s_Equals:
		    other=evaluate(ctxt,obj,var,o.m_Expr2);
		    ret=g_False;
		    if (other!=null){
			boolean equals=false;
			if (item1 instanceof String)
			    equals=((String)item1).equals(other.toString());
			else if (item1 instanceof Integer)
			    equals=((Integer)item1).intValue()==intValue(other);
			else if (item1 instanceof Double)
			    equals=((Double)item1).doubleValue()==floatValue(other);
			else if (item1 instanceof Boolean)
			    equals=((Boolean)item1).booleanValue()==boolValue(other);
			else
			    equals=item1.equals(other);
			if (equals)
			    ret=g_True;
		    }
		    break;
		case NodeTypes.s_NotEquals:
		    other=evaluate(ctxt,obj,var,o.m_Expr2);
		    if (other!=null && item1.equals(other))
			ret=g_False;
		    else
			ret=g_True;
		    break;
		case NodeTypes.s_Concat:{
		    String str1="";
		    if (item1!=null)
			str1=item1.toString();
		    other=evaluate(ctxt,obj,var,o.m_Expr2);
		    String str2="";
		    if (other!=null)
			str2=other.toString();
		    ret=Utils.addStrings(str1,str2);
		    break;
		}
		case NodeTypes.s_Minus:
		    if (item1!=null){
			if (item1 instanceof Double)
			    ret=new Double(-((Double)item1).doubleValue());
			else
			    ret=new Integer(-intValue(item1));
		    }else
			ret=null;
		    break;
		default:{
			Object item2=evaluate(ctxt,obj,var,o.m_Expr2);
			if (item1!=null && item2!=null){
			    boolean useIntegerArith=item1 instanceof Integer;
			    if (useIntegerArith){
				int val1=intValue(item1);
				int val2=intValue(item2);
				switch(n.getType()){
				case NodeTypes.s_More:
				    if (val1>val2)
					ret=g_True;
				    else
					ret=g_False;
				    break;
				case NodeTypes.s_MoreEqual:
				    if (val1>=val2)
					ret=g_True;
				    else
					ret=g_False;
				    break;
				case NodeTypes.s_Less:
				    if (val1<val2)
					ret=g_True;
				    else
					ret=g_False;
				    break;
				case NodeTypes.s_LessEqual:
				    if (val1<=val2)
					ret=g_True;
				    else
					ret=g_False;
				    break;
				case NodeTypes.s_Plus:
				    ret=new Integer(val1+val2);
				    break;
				case NodeTypes.s_Subtract:
				    ret=new Integer(val1-val2);
				    break;
				case NodeTypes.s_Divide:
				    if (val2==0)
					runtimeException(ctxt,Messages.DivideByZero);
				    else
					ret=new Integer((int)((long)val1/(long)val2));
				    break;
				case NodeTypes.s_Mult:
				    ret=new Integer(val1*val2);break;
				case NodeTypes.s_Mod:
				    ret=new Integer(val1%val2);break;
				}
				if (g_TraceExecution) trace("evaluate value1="+val1+" value2="+val2+" result of "+NodeTypes.g_NodeNames[n.getType()]+"="+intValue(ret));
			    }else{
				double val1=floatValue(item1);
				double val2=floatValue(item2);
				switch(n.getType()){
				case NodeTypes.s_More:
				    if (val1>val2)
					ret=g_True;
				    else
					ret=g_False;
				    break;
				case NodeTypes.s_MoreEqual:
				    if (val1>=val2)
					ret=g_True;
				    else
					ret=g_False;
				    break;
				case NodeTypes.s_Less:
				    if (val1<val2)
					ret=g_True;
				    else
					ret=g_False;
				    break;
				case NodeTypes.s_LessEqual:
				    if (val1<=val2)
					ret=g_True;
				    else
					ret=g_False;
				    break;
				case NodeTypes.s_Plus:
				    ret=new Double(val1+val2);break;
				case NodeTypes.s_Subtract:
				    ret=new Double(val1-val2);break;
				case NodeTypes.s_Divide:
				    if (val2==0)
					runtimeException(ctxt,Messages.DivideByZero);
				    else
					ret=new Double(val1/val2);
				    break;
				case NodeTypes.s_Mult:
				    ret=new Double(val1*val2);break;
				case NodeTypes.s_Mod:
				    ret=new Double(val1%val2);break;
				}
				if (g_TraceExecution) trace("evaluate value1="+val1+" value2="+val2+" result of "+NodeTypes.g_NodeNames[n.getType()]+"="+floatValue(ret));
			    }
			}else
			    ret=null;
		    }
		}
	    }
	}
	if (g_TraceExecution) trace("result of "+NodeTypes.g_NodeNames[n.getType()]+"="+((ret!=null)?ret.toString():"null"));
	return ret;
    }
    /**
     * This method parses a string full of Simkin code and returns 
     * the parse tree. If a parse error occurs, a ParseException is thrown
     * <p>For example the following string could be passed:
     * <p>
     <pre>(a,b){
     a=b+1;
     return a;
     }
     </pre>
     * @param location used to help identify the code in error messages
     * @param code a string containing the Simkin code
     * @return a ParseNode representing the parse tree for the code
     */
    public ParseNode parse(String location,String code) throws ParseException{
	StringReader buffer=new StringReader(code);
	Parser parser = new Parser(buffer);
	ParseNode topNode=null;
	try{
	    topNode=parser.executable_code();
	}catch(ParseException e){
	    e.location=location;
	    throw e;
	}catch(TokenMgrError e){
	    ParseException pe=new ParseException(e.getMessage()+" in "+location);
	    pe.location=location;
	    throw pe;
	}
	return topNode;
    }
    /**
     * This method parses a string of Simkin statements, excluding parameters and enclosing braces.
     * The method returns a parse tree. If a parse error occurs, a ParseException is thrown
     * <p> For example the following string could be passed (note the absence of enclosing braces around the statement list):
     <pre>
     a=b+1;
     return a;
     </pre>
     * @param location used to help identify the code in error messages
     * @param paramNames a vector of names for the parameters
     * @param code a string containing the Simkin code
     * @return a ParseNode representing the parse tree for the code
     */
    public ParseNode parseExternalParams(String location,Vector paramNames,String code) throws ParseException{
	StringBufferInputStream buffer=new StringBufferInputStream(code);
	Parser parser = new Parser(buffer);
	MethodDefNode methodNode=new MethodDefNode(0);
	for (int i=0;i<paramNames.size();i++)
	    methodNode.m_Params.addId(new IdNode(0,(String)paramNames.elementAt(i),null,null));
	try{
	    methodNode.m_Stats=parser.stat_list();
	    //	    if (g_TraceExecution) topNode.write(System.out,0);
	}catch(ParseException e){
	    e.location=location;
	    throw e;
	}catch(TokenMgrError e){
	    ParseException pe=new ParseException(e.getMessage()+" in "+location);
	    pe.location=location;
	    throw pe;
	}
	return methodNode;
    }
    /**
     * This method executes a piece of Simkin code within a string
     * <p>For example the following string could be passed:
     * <p>
     <pre>(a,b){
     a=b+1;
     return a;
     }
     </pre>
     * @param location a name used to identify the code in error messages
     * @param obj the object within whose context the code will execute
     * @param code a string containing some Simkin code
     * @param args and array of RValue arguments to the code
     * @return returns an object containing the result of the method and the parse tree for the method, which can be safely cached
     * @throws ParseException this is thrown if there is a syntax error in the code
     * @throws RuntimeException this is thrown if there is an error during the execution of the code (such as method or field not found)
     */
    public ExecuteResult executeString(String location,Object obj,String code,Object [] args) throws RuntimeException, ParseException{
	ParseNode parseNode=parse(location,code);
	Object ret=executeParseTree(location,obj,parseNode,args);
	return new ExecuteResult(ret,parseNode);
    }
    /**
     * This method executes a piece of Simkin code within a string, it assumes that the parameters have been defined outside the statement block.
     * <p> For example the following string could be passed (note the absence of enclosing braces around the statement list):
     <pre>
     a=b+1;
     return a;
     </pre>
     * @param location a name used to identify the code in error messages
     * @param obj the object within whose context the code will execute
     * @param paramNames a vector of names for the parameters
     * @param code a string containing some Simkin code
     * @param args and array of RValue arguments to the code
     * @return returns an object containing the result of the method and the parse tree for the method, which can be safely cached
     * @throws ParseException this is thrown if there is a syntax error in the code
     * @throws RuntimeException this is thrown if there is an error during the execution of the code (such as method or field not found)
     */
    public ExecuteResult executeStringExternalParams(String location,Object obj,Vector paramNames,String code,Object[] args) throws RuntimeException, ParseException{
	ParseNode parseNode=parseExternalParams(location,paramNames,code);
	Object ret=executeParseTree(location,obj,parseNode,args);
	return new ExecuteResult(ret,parseNode);
    }
    /**
     * This method executes a piece of Simkin code from a parse tree.<p>You can call this with a parse tree produced by {@link simkin.Interpreter#parse parse}, {@link simkin.Interpreter#parseExternalParams parseExternalParams},{@link simkin.Interpreter#executeString executeString} or {@link simkin.Interpreter#executeStringExternalParams executeStringExternalParams}. This can be useful for caching parse trees to improve performance.
     * @param location a name used to identify the code in error messages
     * @param obj the object within whose context the code will execute
     * @param parseNode a pre-parsed tree of Simkin code
     * @param args and array of RValue arguments to the code
     * @returns the return value of the method call
     * @throws ParseException this is thrown if there is a syntax error in the code
     * @throws RuntimeException this is thrown if there is an error during the execution of the code (such as method or field not found)
     */
    public Object executeParseTree(String location,Object obj,ParseNode parseNode,Object[] args) throws RuntimeException, ParseException{
	Context ctxt=new Context(location);
	Object ret=null;
	Hashtable pvars=new Hashtable();
	if (parseNode!=null && parseNode.getType()==NodeTypes.s_MethodDefn){
	    MethodDefNode methodNode=(MethodDefNode)parseNode;
	    // fix up parameters by putting them in the local variable table
	    if (args!=null && args.length>0){
		for (int i=0;i<methodNode.m_Params.m_Ids.size();i++){
		    IdNode idNode=(IdNode)(methodNode.m_Params.m_Ids.elementAt(i));
		    String pparam=idNode.m_Name;
		    addLocalVariable(pvars,pparam,args[i]);
		}
	    }
	    ret=executeStats(ctxt,obj,pvars,methodNode.m_Stats);
	}
	return ret;
    }
    /** 
     * this method is called when there is a runtime error and throws a RuntimeException
     */
    static public void runtimeError(String buffer) throws RuntimeException{
	throw new RuntimeException(buffer);
    }
    /** 
     * this method is called when there is a runtime error and throws a RuntimeException, it includes the location in the exception
     */
    void runtimeException(Context ctxt,String buffer) throws RuntimeException{
	if (ctxt!=null){
	    StringBuffer msg=new StringBuffer(ctxt.m_Location);
	    msg.append(": ");
	    msg.append(new Integer(ctxt.m_LineNum).toString());
	    msg.append(": ");
	    msg.append(buffer);
	    throw new RuntimeException(msg.toString());
	}else
	    throw new RuntimeException(buffer);
    }
    /**
     * this method is called when the interpreter wants to output a message to the tracing log
     */
    void traceString(String buffer){
	trace(buffer);
    }
    /*
     * Searches a string for escape sequences \n and \t and converts them
     */
    static String replaceEscapes(String str){
	str=replace(str,"\\n","\n");
	str=replace(str,"\\\"","\"");
	str=replace(str,"\\t","\t");
	return str;
    }
    /**
     * @param original - the string to be replaced
     * @param old_value - the value to be replaced
     * @param new_value - the  new value
     * @return the original string with the old value replaced by the new
     */
    static String replace(String original,String old_value,String new_value){
	String replaced=original;
	int index=original.indexOf(old_value);
	if (index>=0){
	    int pos=0;
	    StringBuffer buffer=new StringBuffer(original.length());
	    while (index>=0){
		buffer.append(original.substring(pos,index));
		buffer.append(new_value);
		pos=index+old_value.length();
		index=original.indexOf(old_value,pos);
	    }
	    if (pos<original.length())
		buffer.append(original.substring(pos));
	    replaced=buffer.toString();
	}
	return replaced;
    }
    /**
     * This method returns the global interpreter associated with all method calls.
     * @return the Interpeter that will be used to execute methods
     */
    public static Interpreter getInterpreter(){
	Interpreter i=null;
	if (g_InterpreterPerThread==false){
	    if (g_Interpreter==null)
		g_Interpreter=new Interpreter();
	    i=g_Interpreter;
	}else{
	    i=(Interpreter)g_Interpreters.get(Thread.currentThread());
	    if (i==null){
		i=new Interpreter();
		g_Interpreters.put(Thread.currentThread(),i);
		if (g_TraceExecution) i.trace("Creating interpreter for thread "+Thread.currentThread());
	    }
	}
	return i;
    }
    /**
     * Sets a global interpreter to be used for method calls - you can use this to override the Interpreter with your own derived classes. If the interpreter-per-thread feature is used, then the interpreter is set only for the current thread.
     * @param the Interpeter that will be used to execute methods - pass null to clear the current value
     */
    public static void setInterpreter(Interpreter i){
	if (g_InterpreterPerThread==false){
	    g_Interpreter=i;
	}else{
	    if (i==null){
		g_Interpreters.remove(Thread.currentThread());
		if (g_TraceExecution) i.trace("Removing interpreter for thread "+Thread.currentThread());
	    }else
		g_Interpreters.put(Thread.currentThread(),i);
	}
    }
    /** This method enables or disables the interpreter-per-thread feature.
     * <p>If the feature is enabled, there is one interpreter for each thread. Calls to getInterpreter() will return the Interpreter defined for the current thread - or create one if one hasn't been encountered yet.
     * <p>The default value for this is true.
     * @param flag true/false to enable or disable the feature
     */
    static public void setInterpreterPerThread(boolean flag){
	g_InterpreterPerThread=flag;
    }
    /** this method outputs a message to the trace - either the global class Tracer, or the current Trace callback */
    void trace(String s){
	if (m_TraceCallback!=null)
	    m_TraceCallback.trace(s);
	else
	    Tracer.trace(s);
    }
    void trace(Exception e){
	if (m_TraceCallback!=null)
	    m_TraceCallback.trace(e.toString());
	else
	    Tracer.trace(e);
    }
    /** this method sets an object to receive trace messages. Pass 0 to fall back to standard mechanism (using Tracer)
     */
    public void setTraceCallback(TraceCallback cb){
	m_TraceCallback=cb;
    }
    /** this method sets an object to be called each time a statement is executed. */
    public void setStatementStepper(StatementStepper stepper){
	m_StatementStepper=stepper;
    }
    /**
     * This function attempts to make a public method call on a Java object by using reflection
     * @param owner - the object to call a method on
     * @param method_name - the name of the method to call
     * @param arguments - the method arguments - types Integer, Double, Boolean and Character are converted to their primitive types
     * @return the result of the method call
     * @throws RuntimeException - if the method could not be called
     * @throws MethodNotSupportedException - if the method could not be found
     */
    public Object reflectiveMethodCall(Context ctxt,Object owner,String method_name,Object[] arguments) throws RuntimeException,MethodNotSupportedException{
	Object ret=null;
	Class[] arg_types=null;
	Class[] arg_types_str=null;
	Class string_class="".getClass();
	if (arguments!=null && arguments.length>0){
	    arg_types=new Class[arguments.length];
	    arg_types_str=new Class[arguments.length];
	    for (int i=0;i<arguments.length;i++){
		if (arguments[i] instanceof Integer){
		    arg_types[i]=Integer.TYPE;
		    arg_types_str[i]=Integer.TYPE;
		}else if (arguments[i] instanceof Double){
		    arg_types[i]=Double.TYPE;
		    arg_types_str[i]=Double.TYPE;
		}else if (arguments[i] instanceof Boolean){
		    arg_types[i]=Boolean.TYPE;
		    arg_types_str[i]=Boolean.TYPE;
		}else{
		    arg_types[i]=arguments[i].getClass();
		    arg_types_str[i]=string_class;
		}
	    }
	}
	if (m_Tracing){
	    StringBuffer args_buffer=new StringBuffer();
	    if (arg_types!=null){
		for (int i=0;i<arg_types.length;i++){
		    args_buffer.append(arg_types[i].getName());
		    args_buffer.append(" ");
		}
	    }
	    trace("reflectiveMethodCall "+method_name+" on "+owner.getClass().getName()+" Argument types:"+args_buffer.toString());
	}
	Method method=null;
	boolean call_with_strings=false;
	try{
	    method=owner.getClass().getMethod(method_name,arg_types);
	}catch(NoSuchMethodException e1){
	    try{
		method=owner.getClass().getMethod(method_name,arg_types_str);
		call_with_strings=true;
	    }catch(NoSuchMethodException e){
		trace(owner.getClass().getName()+"::"+method_name);
		trace(e);
		throw new MethodNotSupportedException(e.toString());
	    }
	}
	catch(Exception e){
	    runtimeException(ctxt,Messages.formatMessage(Messages.CouldNotFindMethod,method_name)+" ("+e.toString()+")");
	}
	if (method!=null){
	    try{
		if (call_with_strings==false){
		    ret=method.invoke(owner,arguments);
		}else{
		    Object [] str_args=new Object[arguments.length];
		    for (int i=0;i<arguments.length;i++){
			if (arg_types_str[i]==string_class)
			    str_args[i]=arguments[i].toString();
			else
			    str_args[i]=arguments[i];
		    }
		    ret=method.invoke(owner,str_args);
		}
		if (ret==null)
		    ret=g_null;
	    }catch(Exception e){
		runtimeException(ctxt,Messages.formatMessage(Messages.CouldNotCallMethod,method_name)+" ("+e.toString()+")");
	    }
	}
	if (g_TraceExecution) trace("reflectiveMethodCall "+method_name+" returns "+(ret!=null?ret.toString():"null"));
	return ret;
    }
    /**
     * This function attempts to set a public Java field using reflection
     * @param owner - the object owning the field
     * @param field_name - the name of the field
     * @param value - the value to assign to the field
     * @throws RuntimeException - if the field could not be set (e.g. if it was private)
     * @throws FieldNotSupportedException - if the field could not be found
     */
    public void reflectiveSetValue(Context ctxt,Object owner,String field_name,Object value) throws RuntimeException,FieldNotSupportedException{
	boolean bFound=false;
	Field field=null;
	Class field_class=null;
	try{
	    field=owner.getClass().getField(field_name);
	    field_class=field.getType();
	    bFound=true;	
	}catch(NoSuchFieldException e){
	    throw new FieldNotSupportedException(e.toString());
	}catch(Exception e){
	    if (g_TraceExecution) trace(e);
	    runtimeException(ctxt,Messages.formatMessage(Messages.CouldNotAccessField,field_name)+" ("+e.toString()+")");
	}
	if (bFound){
	    try{
		field.set(owner,value);
	    }catch(Exception e){
		runtimeException(ctxt,Messages.formatMessage(Messages.CouldNotSetField,field_name)+" ("+e.toString()+")");
	    }
	}
    }
    /**
     * This function attempts to retrieve a public Java field using reflection
     * @param owner - the object owning the field
     * @param field_name - the name of the field
     * @return the value of the field
     * @throws RuntimeException - if the field could not be retrieved (e.g. if it was private)
     * @throws FieldNotSupportedException - if the field could not be found
     */
    public Object reflectiveGetValue(Context ctxt,Object owner,String field_name) throws RuntimeException,FieldNotSupportedException{
	Object fieldValue=null;
	Field field=null;
	try{
	    Class owner_class=owner.getClass();
	    if (owner_class.isArray() && (field_name.equals("length") || field_name.equals("numChildren")))
		fieldValue=new Integer(((Object[])owner).length);
	    else
		field=owner_class.getField(field_name);
	}catch(NoSuchFieldException e){
	    throw new FieldNotSupportedException(e.toString());
	}catch(Exception e){
	    runtimeException(ctxt,Messages.formatMessage(Messages.CouldNotAccessField,field_name)+" ("+e.toString()+")");
	}
	if (fieldValue==null){
	    try{
		fieldValue=field.get(owner);
	    }catch(Exception e){
		runtimeException(ctxt,Messages.formatMessage(Messages.CouldNotReadField,field_name)+" ("+e.toString()+")");
	    }
	    if (fieldValue==null)
		fieldValue=g_null;
	}
	if (g_TraceExecution) trace("reflectiveGetValue: on "+owner.getClass().getName()+" field name: "+field_name+" returns: "+fieldValue.getClass().getName());
	return fieldValue;
    }
};
